package xgs.net.searcher;

import xgs.net.utils.ClassUtils;
import xgs.net.utils.IOUtils;
import xgs.net.utils.URLUtils;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.JarURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.util.Collection;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

public abstract class FileSearcher
{
    public void lookupFileSystem(final File dir) {
        this.doLookupInFileSystem(dir, null, null);
    }
    
    public void lookupClasspath(final String... packageNames) {
        if (packageNames == null || packageNames.length == 0) {
            final Collection<URL> urls = ClassUtils.getClasspathURLs((String)null);
            this.doGetClasspathResources(urls, null);
        }
        else {
            for (final String pkg : packageNames) {
                final Collection<URL> urls2 = ClassUtils.getClasspathURLs(pkg);
                this.doGetClasspathResources(urls2, pkg.replace('.', '/'));
            }
        }
    }
    
    public void lookupZipFile(final File zipFile, final String entryName) {
        ZipFile zip = null;
        try {
            zip = new ZipFile(zipFile);
            this.doLookupInZipFile(zip, entryName);
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
        finally {
            IOUtils.closeQuietly(zip);
        }
    }
    
    private void doGetClasspathResources(final Collection<URL> urls, final String pkgdir) {
        for (final URL url : urls) {
            final String protocol = url.getProtocol();
            System.out.println(protocol + " \t " + url);
            if ("file".equals(protocol)) {
                final File file = URLUtils.toFileObject(url);
                if (file.isDirectory()) {
                    this.doLookupInFileSystem(file, pkgdir, null);
                }
                else {
                    final String name = file.getName().toLowerCase();
                    if (!name.endsWith(".jar") && !name.endsWith(".zip")) {
                        continue;
                    }
                    this.doLookupInZipFile(url, pkgdir);
                }
            }
            else if ("jar".equals(protocol) || "zip".equals(protocol)) {
                this.doLookupInZipFile(url, pkgdir);
            }
            else {
                if (!"vfs".equals(protocol)) {
                    throw new IllegalStateException("Unsupported url format: " + url.toString());
                }
                this.doLookupInVfsFile(url, pkgdir);
            }
        }
    }
    
    private void doLookupInFileSystem(final File dir, final String pkgdir, final String relativeName) {
        if (!dir.exists() || !dir.isDirectory()) {
            return;
        }
        final File[] files = dir.listFiles();
        if (files == null) {
            return;
        }
        for (final File file : files) {
            final String name = (relativeName == null) ? file.getName() : (relativeName + '/' + file.getName());
            final SystemFileEntry entry = new SystemFileEntry(file, pkgdir, name);
            if (file.isDirectory()) {
                if (this.visitSystemDirEntry(entry)) {
                    this.doLookupInFileSystem(file, pkgdir, name);
                }
            }
            else {
                this.visitSystemFileEntry(entry);
            }
        }
    }
    
    private void doLookupInZipFile(final URL url, final String pkgdir) {
        ZipFile zip = null;
        try {
            if ("jar".equals(url.getProtocol())) {
                zip = ((JarURLConnection)url.openConnection()).getJarFile();
            }
            else {
                final File file = URLUtils.toFileObject(url);
                if (!file.exists()) {
                    return;
                }
                zip = new ZipFile(file);
            }
            this.doLookupInZipFile(zip, pkgdir);
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
        finally {
            IOUtils.closeQuietly(zip);
        }
    }
    
    private void doLookupInZipFile(final ZipFile zip, String pkgdir) {
        if (pkgdir == null || pkgdir.length() == 0) {
            pkgdir = null;
        }
        else {
            pkgdir += '/';
        }
        final Enumeration<? extends ZipEntry> entries = zip.entries();
        while (entries.hasMoreElements()) {
            final ZipEntry entry = (ZipEntry)entries.nextElement();
            String entryName = entry.getName();
            if (entry.isDirectory()) {
                entryName = entryName.substring(0, entryName.length() - 1);
            }
            if (pkgdir == null) {
                if (entry.isDirectory()) {
                    this.visitZipDirEntry(new ZipFileEntry(zip, entry, entryName));
                }
                else {
                    this.visitZipFileEntry(new ZipFileEntry(zip, entry, entryName));
                }
            }
            else {
                if (!entryName.startsWith(pkgdir)) {
                    continue;
                }
                entryName = entryName.substring(pkgdir.length());
                if (entry.isDirectory()) {
                    this.visitZipDirEntry(new ZipFileEntry(zip, entry, entryName));
                }
                else {
                    this.visitZipFileEntry(new ZipFileEntry(zip, entry, entryName));
                }
            }
        }
    }
    
    private void doLookupInVfsFile(final URL url, final String pkgdir) {
        try {
            final URLConnection conn = url.openConnection();
            if (!conn.getClass().getName().equals("org.jboss.vfs.protocol.VirtualFileURLConnection")) {
                throw new IllegalStateException("Unsupported URL: " + url);
            }
            final String vfs = conn.getContent().toString();
            final File file = new File(vfs.substring(1, vfs.length() - 1));
            if (!file.exists()) {
                return;
            }
            if (file.isDirectory()) {
                this.doLookupInFileSystem(file, pkgdir, null);
            }
            else {
                final String name = file.getName().toLowerCase();
                if (name.endsWith(".jar") || name.endsWith(".zip")) {
                    final ZipFile zip = new ZipFile(file);
                    try {
                        this.doLookupInZipFile(zip, pkgdir);
                    }
                    finally {
                        IOUtils.closeQuietly(zip);
                    }
                }
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
    
    protected boolean visitSystemDirEntry(final SystemFileEntry dir) {
        return this.visitDirEntry(dir);
    }
    
    protected void visitSystemFileEntry(final SystemFileEntry file) {
        this.visitFileEntry(file);
    }
    
    protected void visitZipDirEntry(final ZipFileEntry dir) {
        this.visitDirEntry(dir);
    }
    
    protected void visitZipFileEntry(final ZipFileEntry file) {
        this.visitFileEntry(file);
    }
    
    protected boolean visitDirEntry(final FileEntry dir) {
        return true;
    }
    
    protected void visitFileEntry(final FileEntry file) {
    }
    
    public static class SystemFileEntry implements FileEntry
    {
        private final File file;
        private final String pkgdir;
        private final String relativeName;
        
        public SystemFileEntry(final File file, final String pkgdir, final String relativeName) {
            this.file = file;
            this.pkgdir = pkgdir;
            this.relativeName = relativeName;
        }
        
        public File getFile() {
            return this.file;
        }
        
        @Override
        public boolean isDirectory() {
            return this.file.isDirectory();
        }
        
        @Override
        public boolean isJavaClass() {
            return !this.file.isDirectory() && this.file.getName().endsWith(".class");
        }
        
        @Override
        public String getName() {
            return this.file.getName();
        }
        
        @Override
        public String getRelativePathName() {
            return this.relativeName;
        }
        
        @Override
        public String getQualifiedFileName() {
            String name;
            if (this.pkgdir != null) {
                name = this.pkgdir + '/' + this.relativeName;
            }
            else {
                name = this.relativeName;
            }
            if (this.file.isDirectory()) {
                return name.replace('/', '.');
            }
            return name;
        }
        
        @Override
        public String getQualifiedJavaName() {
            final String name = this.getQualifiedFileName();
            if (this.file.isDirectory()) {
                return name;
            }
            if (name.endsWith(".class")) {
                return name.substring(0, name.length() - 6).replace('/', '.');
            }
            throw new IllegalStateException("FileEntry is not a Java Class: " + this.toString());
        }
        
        @Override
        public long length() {
            return this.file.length();
        }
        
        @Override
        public long lastModified() {
            return this.file.lastModified();
        }
        
        @Override
        public InputStream getInputStream() throws IOException {
            return new FileInputStream(this.file);
        }
        
        @Override
        public String toString() {
            return this.file.toString();
        }
    }
    
    public static class ZipFileEntry implements FileEntry
    {
        private final ZipFile zip;
        private final ZipEntry entry;
        private final String relativeName;
        
        public ZipFileEntry(final ZipFile zip, final ZipEntry entry, final String relativeName) {
            this.zip = zip;
            this.entry = entry;
            this.relativeName = relativeName;
        }
        
        public ZipFile getZipFile() {
            return this.zip;
        }
        
        public ZipEntry getZipEntry() {
            return this.entry;
        }
        
        @Override
        public boolean isDirectory() {
            return this.entry.isDirectory();
        }
        
        @Override
        public boolean isJavaClass() {
            return this.entry.getName().endsWith(".class");
        }
        
        @Override
        public String getName() {
            final int ipos = this.relativeName.lastIndexOf(47);
            return (ipos != -1) ? this.relativeName.substring(ipos + 1) : this.relativeName;
        }
        
        @Override
        public String getRelativePathName() {
            return this.relativeName;
        }
        
        @Override
        public String getQualifiedFileName() {
            final String name = this.entry.getName();
            if (this.entry.isDirectory()) {
                return name.substring(0, name.length() - 1).replace('/', '.');
            }
            return name;
        }
        
        @Override
        public String getQualifiedJavaName() {
            final String name = this.getQualifiedFileName();
            if (this.entry.isDirectory()) {
                return name;
            }
            if (name.endsWith(".class")) {
                return name.substring(0, name.length() - 6).replace('/', '.');
            }
            throw new IllegalStateException("FileEntry is not a Java Class: " + this.toString());
        }
        
        @Override
        public long length() {
            return this.entry.getSize();
        }
        
        @Override
        public long lastModified() {
            return this.entry.getTime();
        }
        
        @Override
        public InputStream getInputStream() throws IOException {
            return this.zip.getInputStream(this.entry);
        }
        
        @Override
        public String toString() {
            return this.entry.toString();
        }
    }
    
    public interface FileEntry
    {
        boolean isDirectory();
        
        boolean isJavaClass();
        
        String getName();
        
        String getRelativePathName();
        
        String getQualifiedFileName();
        
        String getQualifiedJavaName();
        
        long length();
        
        long lastModified();
        
        InputStream getInputStream() throws IOException;
    }
}
