package xgs.net.utils;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import java.io.*;
import java.util.HashMap;
import java.util.Map;

public class XmlHelper
{
    private final XPath path;
    private final Document doc;
    
    private XmlHelper(final InputSource inputSource) throws ParserConfigurationException, SAXException, IOException {
        final DocumentBuilderFactory dbf = getDocumentBuilderFactory();
        final DocumentBuilder db = dbf.newDocumentBuilder();
        this.doc = db.parse(inputSource);
        this.path = getXPathFactory().newXPath();
    }
    
    private static XmlHelper create(final InputSource inputSource) {
        try {
            return new XmlHelper(inputSource);
        }
        catch (ParserConfigurationException e) {
            throw new RuntimeException(e);
        }
        catch (SAXException e2) {
            throw new RuntimeException(e2);
        }
        catch (IOException e3) {
            throw new RuntimeException(e3);
        }
    }
    
    public static XmlHelper of(final InputStream is) {
        final InputSource inputSource = new InputSource(is);
        return create(inputSource);
    }
    
    public static XmlHelper of(final File file) {
        FileInputStream fis = null;
        try {
            fis = new FileInputStream(file);
            return of(fis);
        }
        catch (FileNotFoundException e) {
            throw new RuntimeException(e);
        }
        finally {
            IOUtils.closeQuietly(fis);
        }
    }
    
    public static XmlHelper of(final String xmlStr) {
        final StringReader sr = new StringReader(xmlStr.trim());
        final InputSource inputSource = new InputSource(sr);
        final XmlHelper xmlHelper = create(inputSource);
        IOUtils.closeQuietly(sr);
        return xmlHelper;
    }
    
    private Object evalXPath(final String expression, Object item, final QName returnType) {
        item = ((null == item) ? this.doc : item);
        try {
            return this.path.evaluate(expression, item, returnType);
        }
        catch (XPathExpressionException e) {
            throw new RuntimeException(e);
        }
    }
    
    public String getString(final String expression) {
        return (String)this.evalXPath(expression, null, XPathConstants.STRING);
    }
    
    public Boolean getBoolean(final String expression) {
        return (Boolean)this.evalXPath(expression, null, XPathConstants.BOOLEAN);
    }
    
    public Number getNumber(final String expression) {
        return (Number)this.evalXPath(expression, null, XPathConstants.NUMBER);
    }
    
    public Node getNode(final String expression) {
        return (Node)this.evalXPath(expression, null, XPathConstants.NODE);
    }
    
    public NodeList getNodeList(final String expression) {
        return (NodeList)this.evalXPath(expression, null, XPathConstants.NODESET);
    }
    
    public String getString(final Object node, final String expression) {
        return (String)this.evalXPath(expression, node, XPathConstants.STRING);
    }
    
    public Boolean getBoolean(final Object node, final String expression) {
        return (Boolean)this.evalXPath(expression, node, XPathConstants.BOOLEAN);
    }
    
    public Number getNumber(final Object node, final String expression) {
        return (Number)this.evalXPath(expression, node, XPathConstants.NUMBER);
    }
    
    public Node getNode(final Object node, final String expression) {
        return (Node)this.evalXPath(expression, node, XPathConstants.NODE);
    }
    
    public NodeList getNodeList(final Object node, final String expression) {
        return (NodeList)this.evalXPath(expression, node, XPathConstants.NODESET);
    }
    
    public Map<String, String> toMap() {
        final Element root = this.doc.getDocumentElement();
        final Map<String, String> params = new HashMap<String, String>();
        final NodeList list = root.getChildNodes();
        for (int i = 0; i < list.getLength(); ++i) {
            final Node node = list.item(i);
            if (node instanceof Element) {
                params.put(node.getNodeName(), node.getTextContent());
            }
        }
        return params;
    }
    
    public void outXML(final OutputStream out) {
        PrintWriter writer = null;
        try {
            writer = new PrintWriter(new OutputStreamWriter(out, Charsets.UTF_8));
            final StreamResult result = new StreamResult(writer);
            this.outXML(result);
        }
        catch (TransformerException e) {
            throw new RuntimeException(e);
        }
        finally {
            IOUtils.closeQuietly(writer);
        }
    }
    
    private void outXML(final StreamResult result) throws TransformerException {
        final DOMSource source = new DOMSource(this.doc);
        final Transformer transformer = getTransFactory().newTransformer();
        transformer.setOutputProperty("indent", "yes");
        transformer.setOutputProperty("omit-xml-declaration", "yes");
        transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
        transformer.transform(source, result);
    }
    
    private static DocumentBuilderFactory getDocumentBuilderFactory() {
        return XmlHelperHolder.documentBuilderFactory;
    }
    
    private static XPathFactory getXPathFactory() {
        return XmlHelperHolder.xPathFactory;
    }
    
    private static TransformerFactory getTransFactory() {
        return XmlHelperHolder.transFactory;
    }
    
    private static class XmlHelperHolder
    {
        private static DocumentBuilderFactory documentBuilderFactory;
        private static XPathFactory xPathFactory;
        private static TransformerFactory transFactory;
        
        static {
            XmlHelperHolder.documentBuilderFactory = DocumentBuilderFactory.newInstance();
            XmlHelperHolder.xPathFactory = XPathFactory.newInstance();
            XmlHelperHolder.transFactory = TransformerFactory.newInstance();
        }
    }
}
